@@ -26,6 +26,7 @@
 #include <errno.h>
 #include <sys/wait.h>
 #include <string.h>
+#include <stdlib.h>
 
 #include <glib/gi18n.h>
 #include <gtk/gtk.h>
@@ -101,6 +102,8 @@ struct GSWindowPrivate
         guint      info_bar_timer_id;
         guint      clock_update_id;
 
+        gint64     clock_update_time;
+
         gint       lock_pid;
         gint       lock_watch_id;
         gint       dialog_response;
@@ -264,12 +267,13 @@ gs_window_clear_to_background_surface (GSWindow *window)
 static void
 clear_widget (GtkWidget *widget)
 {
-        GdkRGBA rgba = { 0.0, 0.0, 0.0, 1.0 };
+        GdkColor color = { 0, 0x0000, 0x0000, 0x0000 };
 
         if (!gtk_widget_get_realized (widget))
                 return;
 
-        gtk_widget_override_background_color (widget, GTK_STATE_FLAG_NORMAL, &rgba);
+        gdk_window_set_background (gtk_widget_get_window (widget),
+                                   &color);
         gtk_widget_queue_draw (GTK_WIDGET (widget));
 }
 
@@ -480,9 +484,9 @@ static void
 add_watchdog_timer (GSWindow *window,
                     glong     timeout)
 {
-        window->priv->watchdog_timer_id = g_timeout_add (timeout,
-                                                         (GSourceFunc)watchdog_timer,
-                                                         window);
+        window->priv->watchdog_timer_id = g_timeout_add_seconds (timeout,
+                                                                 (GSourceFunc)watchdog_timer,
+                                                                 window);
 }
 
 static void
@@ -670,7 +674,7 @@ gs_window_real_show (GtkWidget *widget)
         window->priv->timer = g_timer_new ();
 
         remove_watchdog_timer (window);
-        add_watchdog_timer (window, 30000);
+        add_watchdog_timer (window, 30);
 
         select_popup_events ();
         window_select_shape_events (window);
@@ -1116,9 +1120,7 @@ keyboard_command_finish (GSWindow *window)
         gs_debug ("Keyboard finished");
 
         if (window->priv->keyboard_pid > 0) {
-                int exit_status;
-
-                exit_status = wait_on_child (window->priv->keyboard_pid);
+                wait_on_child (window->priv->keyboard_pid);
 
                 g_spawn_close_pid (window->priv->keyboard_pid);
                 window->priv->keyboard_pid = 0;
@@ -1247,9 +1249,7 @@ gs_window_dialog_finish (GSWindow *window)
         kill_dialog_command (window);
 
         if (window->priv->lock_pid > 0) {
-                int exit_status;
-
-                exit_status = wait_on_child (window->priv->lock_pid);
+                wait_on_child (window->priv->lock_pid);
 
                 g_spawn_close_pid (window->priv->lock_pid);
                 window->priv->lock_pid = 0;
@@ -2200,7 +2200,8 @@ update_clock (GSWindow *window)
                 /* Translators, this is the 12h date format used in the panel clock */
                 clock_format = _("%a %l:%M %p");
 
-        dt = g_date_time_new_now_local ();
+        window->priv->clock_update_time = g_get_real_time ();
+        dt = g_date_time_new_from_unix_local (window->priv->clock_update_time / G_USEC_PER_SEC);
         text = g_date_time_format (dt, clock_format);
         markup = g_strdup_printf ("<b><span foreground=\"#ccc\">%s</span></b>", text);
         gtk_label_set_markup (GTK_LABEL (window->priv->clock), markup);
@@ -2218,6 +2219,28 @@ update_clock_timer (GSWindow *window)
         return FALSE;
 }
 
+static gboolean
+check_clock_timer (GSWindow *window)
+{
+        /* Update the panel clock when necessary.
+           This happens:
+
+           - Once a minute in the normal case
+           - When the machine resumes from suspend
+           - When the system time is adjusted
+
+           Right now this function is called much more frequently than any of the
+           above 3 events happen (see queue_clock_update ()).
+
+           We can wake up less often if bug 655129 gets fixed.  */
+        if (ABS (g_get_real_time () - window->priv->clock_update_time) > 60 * G_USEC_PER_SEC) {
+                update_clock (window);
+        }
+
+        queue_clock_update (window);
+        return FALSE;
+}
+
 static void
 queue_clock_update (GSWindow *window)
 {
@@ -2227,10 +2250,17 @@ queue_clock_update (GSWindow *window)
         gettimeofday (&tv, NULL);
         timeouttime = (G_USEC_PER_SEC - tv.tv_usec) / 1000 + 1;
 
-        /* timeout of one minute if we don't care about the seconds */
+        /* time until next minute */
         timeouttime += 1000 * (59 - tv.tv_sec % 60);
 
-        window->priv->clock_update_id = g_timeout_add (timeouttime, (GSourceFunc)update_clock_timer, window);
+        /* If we are more than 2.5 seconds from the start of the next minute,
+           schedule less precise but more power friendly 2 second add_seconds
+           timeout to check if the system realtime clock has changed under us. */
+        if (timeouttime > 2500) {
+                window->priv->clock_update_id = g_timeout_add_seconds (2, (GSourceFunc)check_clock_timer, window);
+        } else {
+                window->priv->clock_update_id = g_timeout_add (timeouttime, (GSourceFunc)update_clock_timer, window);
+        }
 }
 
 static char *
